from __future__ import annotations

import copy
from pathlib import Path
from typing import Dict, Optional, Union

import yaml

from cyberdrop_dl.base_functions.base_functions import log
from cyberdrop_dl.base_functions.config_schema import config_default


def _to_config_value(value) -> Union[str, int]:
    return str(value) if isinstance(value, Path) else value


def _create_config(config: Path, passed_args: Optional[dict] = None, enabled=False) -> Dict:
    """Creates the default config file, or remakes it with passed arguments"""
    config_data: Dict = config_default
    if passed_args:
        config_data["Apply_Config"] = enabled
        for group in config_data["Configuration"].values():
            for arg in group:
                if arg in passed_args:
                    group[arg] = _to_config_value(passed_args[arg])

    with open(config, 'w') as yamlfile:
        yaml.dump(config_data, yamlfile)

    return config_data


def _validate_config(config: Path) -> Dict:
    """Validates the existing config file"""
    with open(config, "r") as yamlfile:
        config_data = yaml.load(yamlfile, Loader=yaml.FullLoader)
    try:
        data = config_data["Configuration"]
        enabled = config_data["Apply_Config"]

        config_groups: Dict = config_default["Configuration"]
        if all(set(group) <= set(data[group_name]) for group_name, group in config_groups.items()):
            return config_data

        config.unlink()

        args = {}
        for group_name in config_groups:
            args.update(data[group_name])
        config_data = _create_config(config, args, enabled)

    except (KeyError, TypeError):
        config.unlink()
        config_data = _create_config(config)

    return config_data


def run_args(config: Path, cmd_arg: Dict) -> Dict:
    """Returns the proper runtime arguments based on the config and command line arguments"""
    data = _validate_config(config) if config.is_file() else _create_config(config, cmd_arg)
    if data['Apply_Config']:
        log("Running using config arguments")
        data = data["Configuration"]
        for file, path in data['Files'].items():
            data['Files'][file] = Path(path)
        data['Sorting']['sort_directory'] = Path(data['Sorting']['sort_directory'])
        return data
    else:
        log("Running using command line arguments. If you are trying to use the config file, change 'Apply_Config' to True")

    config_data: Dict = config_default["Configuration"]
    for group in config_data.values():
        for arg in group:
            if arg in cmd_arg:
                group[arg] = cmd_arg[arg]
    return config_data


async def document_args(args: Dict) -> None:
    """We document the runtime arguments for debugging and troubleshooting, redacting sensitive information"""
    print_args = copy.deepcopy(args)

    log("Starting Cyberdrop-DL")
    for group_name, group in print_args.items():
        args_type = group_name.replace('_', ' ').lower()
        for arg in group:
            if group[arg] is not None and any(s in arg for s in ('api_key', 'password', 'imgur', 'reddit', 'cookie')):
                group[arg] = '!REDACTED!'
        log(f"Using {args_type} arguments: {group}", quiet=True)



from __future__ import annotations

import html
import re
from typing import TYPE_CHECKING

from aiolimiter import AsyncLimiter
from yarl import URL

from ..base_functions.base_functions import (
    FILE_FORMATS,
    get_filename_and_ext,
    log,
    logger,
    make_title_safe,
)
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure

if TYPE_CHECKING:
    from ..base_functions.base_functions import ErrorFileWriter
    from ..base_functions.sql_helper import SQLHelper
    from ..client.client import ScrapeSession


class BunkrCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper, remove_bunkr_id: bool, error_writer: ErrorFileWriter):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper
        self.remove_bunkr_id = remove_bunkr_id
        self.limiter = AsyncLimiter(10, 1)
        self.small_limiter = AsyncLimiter(1, 1)

        self.error_writer = error_writer

        self.primary_base_domain = URL("https://bunkrr.su")
        self.api_link = URL(f"https://api-v2.{self.primary_base_domain.host}")

    async def fetch(self, session: ScrapeSession, url: URL) -> AlbumItem:
        """Scraper for Bunkr"""
        album_obj = AlbumItem("Loose Bunkr Files", [])
        log(f"Starting: {url}", quiet=self.quiet, style="green")

        url = await self.get_stream_link(url)

        if "a" in url.parts:
            album_obj = await self.get_album(session, url)
            await self.SQL_Helper.insert_album("bunkr", url, album_obj)

            if album_obj.media:
                log(f"Finished: {url}", quiet=self.quiet, style="green")
            return album_obj

        elif "v" in url.parts:
            media = await self.get_video(session, url)
        else:
            media = await self.get_other(session, url)
        if not media.filename:
            return album_obj
        await album_obj.add_media(media)
        log(f"Finished: {url}", quiet=self.quiet, style="green")
        if not media.complete:
            await self.SQL_Helper.insert_media("bunkr", "", media)
        return album_obj

    async def get_stream_link(self, url: URL):
        cdn_possibilities = r"^(?:(?:(?:media-files|cdn|c|pizza|cdn-burger)[0-9]{0,2})|(?:(?:big-taco-|cdn-pizza)[0-9]{0,2}(?:redir)?))\.bunkr?\.[a-z]{2,3}$"

        if not re.match(cdn_possibilities, url.host):
            return url

        ext = url.suffix.lower()
        if ext == "":
            return url

        if ext in FILE_FORMATS['Images']:
            url = self.primary_base_domain / "i" / url.parts[-1]
        elif ext in FILE_FORMATS['Videos']:
            url = self.primary_base_domain / "v" / url.parts[-1]
        else:
            url = self.primary_base_domain / "d" / url.parts[-1]

        return url

    async def remove_id(self, filename: str, ext: str):
        """Removes the additional string bunkr adds to the end of every filename"""
        original_filename = filename
        if self.remove_bunkr_id:
            filename = filename.rsplit(ext, 1)[0]
            filename = filename.rsplit("-", 1)[0]
            if ext not in filename:
                filename = filename + ext
        return original_filename, filename

    async def get_video(self, session: ScrapeSession, url: URL):
        try:
            async with self.limiter:
                soup = await session.get_BS4(url)
                link = soup.select("a[class*=bg-blue-500]")
                link = link[-1]
                link_resp = URL(link.get("href"))

            try:
                filename, ext = await get_filename_and_ext(link_resp.name)
            except NoExtensionFailure:
                filename, ext = await get_filename_and_ext(url.name)

            original_filename, filename = await self.remove_id(filename, ext)

            await self.SQL_Helper.fix_bunkr_entries(link_resp, original_filename)
            complete = await self.SQL_Helper.check_complete_singular("bunkr", link_resp)
            return MediaItem(link_resp, url, complete, filename, ext, original_filename)
        except Exception as e:
            logger.debug("Error encountered while handling %s", url, exc_info=True)
            await self.error_writer.write_errored_scrape(url, e, self.quiet)
            logger.debug(e)
            return MediaItem(url, url, False, "", "", "")

    async def get_other(self, session: ScrapeSession, url: URL):
        try:
            async with self.limiter:
                soup = await session.get_BS4(url)
                link = soup.select('a[class*="text-white inline-flex"]')
                link = link[-1]
                link_resp = URL(link.get("href"))

            try:
                filename, ext = await get_filename_and_ext(link_resp.name)
            except NoExtensionFailure:
                filename, ext = await get_filename_and_ext(url.name)

            original_filename, filename = await self.remove_id(filename, ext)

            await self.SQL_Helper.fix_bunkr_entries(link_resp, original_filename)
            complete = await self.SQL_Helper.check_complete_singular("bunkr", link_resp)
            return MediaItem(link_resp, url, complete, filename, ext, original_filename)
        except Exception as e:
            logger.debug("Error encountered while handling %s", url, exc_info=True)
            await self.error_writer.write_errored_scrape(url, e, self.quiet)
            logger.debug(e)
            return MediaItem(url, url, False, "", "", "")

    async def get_album(self, session: ScrapeSession, url: URL):
        """Iterates through an album and creates the media items"""

        url = self.primary_base_domain.with_path(url.path)

        album = AlbumItem(url.name, [])
        try:
            async with self.limiter:
                soup = await session.get_BS4(url)
            title = soup.select_one('h1[class="text-[24px] font-bold text-dark dark:text-white"]')
            for elem in title.find_all("span"):
                elem.decompose()
            title = await make_title_safe(title.get_text())
            await album.set_new_title(title)
            for file in soup.select('a[class*="grid-images_box-link"]'):
                link = file.get("href")

                assert url.host is not None
                if link.startswith("/"):
                    link = URL("https://" + url.host + link)
                link = URL(link)

                try:
                    referer = await self.get_stream_link(link)
                except Exception as e:
                    logger.debug("Error encountered while handling %s", link, exc_info=True)
                    await self.error_writer.write_errored_scrape(link, e, self.quiet)
                    logger.debug(e)
                    continue

                if "v" in referer.parts:
                    media = await self.get_video(session, referer)
                else:
                    media = await self.get_other(session, referer)
                if not media.filename:
                    continue
                await album.add_media(media)

        except Exception as e:
            logger.debug("Error encountered while handling %s", url, exc_info=True)
            await self.error_writer.write_errored_scrape(url, e, self.quiet)

        return album



from __future__ import annotations

import itertools
from http import HTTPStatus
from typing import TYPE_CHECKING, List

from yarl import URL

from ..base_functions.base_functions import create_media_item, log, logger
from ..base_functions.data_classes import AlbumItem, MediaItem
from ..base_functions.error_classes import NoExtensionFailure

if TYPE_CHECKING:
    from ..base_functions.base_functions import ErrorFileWriter
    from ..base_functions.sql_helper import SQLHelper
    from ..client.client import ScrapeSession


class PostImgCrawler:
    def __init__(self, quiet: bool, SQL_Helper: SQLHelper, error_writer: ErrorFileWriter):
        self.quiet = quiet
        self.SQL_Helper = SQL_Helper

        self.error_writer = error_writer

    async def fetch(self, session: ScrapeSession, url: URL) -> AlbumItem:
        """Director for PostImg scraping"""
        album_obj = AlbumItem("Loose PostImg Files", [])
        log(f"Starting: {url}", quiet=self.quiet, style="green")

        try:
            if "gallery" in url.parts:
                content = await self.get_folder(session, url)
                for media_item in content:
                    await album_obj.add_media(media_item)
            elif url.host == "i.postimg.cc":
                url = URL("https://postimg.cc/") / url.parts[1]
                media_item = await self.get_singular(session, url)
                await album_obj.add_media(media_item)
            else:
                media_item = await self.get_singular(session, url)
                await album_obj.add_media(media_item)

        except Exception as e:
            logger.debug("Error encountered while handling %s", url, exc_info=True)
            await self.error_writer.write_errored_scrape(url, e, self.quiet)

        await self.SQL_Helper.insert_album("postimg", url, album_obj)
        log(f"Finished: {url}", quiet=self.quiet, style="green")
        return album_obj

    async def get_folder(self, session: ScrapeSession, url: URL) -> List:
        """Handles folder scraping"""
        album = url.raw_name
        data = {"action": "list", "album": album, "page": 0}
        content = []
        for i in itertools.count(1):
            data["page"] = i
            data_out = await session.post(URL("https://postimg.cc/json"), data)
            if data_out['status_code'] != HTTPStatus.OK or not data_out['images']:
                break
            for item in data_out['images']:
                referer = URL("https://postimg.cc/" + item[0])
                img = URL(item[4].replace(item[0], item[1]))

                try:
                    media_item = await create_media_item(img, referer, self.SQL_Helper, "postimg")
                except NoExtensionFailure:
                    logger.debug("Couldn't get extension for %s", img)
                    continue

                content.append(media_item)
        return content

    async def get_singular(self, session: ScrapeSession, url: URL) -> MediaItem:
        """Handles singular folder scraping"""
        soup = await session.get_BS4(url)
        link = URL(soup.select_one("a[id=download]").get('href').replace("?dl=1", ""))
        return await create_media_item(link, url, self.SQL_Helper, "postimg")
